/**
 *  Copyright 2012 Martynas Jusevičius <martynas@graphity.org>
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
package org.graphity.update;

import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.query.Syntax;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.update.UpdateFactory;
import com.hp.hpl.jena.update.UpdateRequest;
import java.io.ByteArrayOutputStream;
import org.graphity.server.update.UpdateProcessRemote;
import org.graphity.query.QueryEngineHTTP;
import org.openjena.riot.WebContent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * SPARQL Update implementation of Fuseki DatasetAccessor
 * @see <a href="http://jena.apache.org/documentation/javadoc/fuseki/org/apache/jena/fuseki/DatasetAccessor.html">Fuseki's DatasetAccessor</a>
 * @author Martynas Jusevičius <martynas@graphity.org>
 */
public class SPARQLAdapter implements org.apache.jena.fuseki.DatasetAccessor
{
    private static final Logger log = LoggerFactory.getLogger(SPARQLAdapter.class);
    
    private String endpoint = null;
    
    /**
     * Constructs adapter based on SPARQL endpoint URI.
     * @param	endpoint    Absolute SPARQL endpoint URI
     */
    public SPARQLAdapter(String endpoint)
    {
	this.endpoint = endpoint;
    }

    /**
     * Returns SPARQL endpoint URI for this adapter.
     * @return	absolute SPARQL endpoint URI
     */
    public String getEndpoint()
    {
	return endpoint;
    }
    
    /**
     * Adds RDF Model to the default graph.
     * @param	data	RDF Model
     */
    @Override
    public void add(Model model)
    {
	ByteArrayOutputStream baos = new ByteArrayOutputStream();
	model.write(baos, WebContent.langNTriples);

	//UpdateDataInsert;
	// http://www.w3.org/TR/sparql11-update/#insertData
	UpdateRequest request = UpdateFactory.create("INSERT DATA { "
	    + baos.toString() +
	    "}", Syntax.syntaxSPARQL_11);

	UpdateProcessRemote process = new UpdateProcessRemote(request, getEndpoint());
	process.execute();	
    }
    
    /**
     * Adds RDF Model to specified named graph.
     * @param	graphUri    URI of the named graph
     * @param	model	    RDF Model
     */
    @Override
    public void add(String graphUri, Model model)
    {
	ByteArrayOutputStream baos = new ByteArrayOutputStream();
	model.write(baos, WebContent.langNTriples);

	//UpdateDataInsert;
	// http://www.w3.org/TR/sparql11-update/#insertData
	UpdateRequest request = UpdateFactory.create("CREATE GRAPH <" + graphUri + ">", Syntax.syntaxSPARQL_11).
	   add("INSERT DATA { GRAPH <" + graphUri + "> {"
	    + baos.toString() +
	    "} }");

	UpdateProcessRemote process = new UpdateProcessRemote(request, getEndpoint());
	process.execute();
    }
    
    /**
     * Checks if named graph exists.
     * @param	graphUri    URI of the named graph
     * @return	true if named graph exists
     */
    @Override
    public boolean containsModel(String graphUri)
    {
	Query query = QueryFactory.create("ASK { <" + graphUri + "> ?p ?o }");
	QueryEngineHTTP request = new QueryEngineHTTP(getEndpoint(), query);
	return request.execAsk();
    }

    @Override
    public Model getModel()
    {
	throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public Model getModel(String string)
    {
	throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void putModel(Model model)
    {
	throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void putModel(String string, Model model)
    {
	throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void deleteDefault()
    {
	throw new UnsupportedOperationException("Not supported yet.");
    }

    @Override
    public void deleteModel(String string)
    {
	throw new UnsupportedOperationException("Not supported yet.");
    }
    
}